# read result file

from itertools import groupby
from collections import defaultdict


def combine(lst):
    # Merge consecutive identical values
    pos = (j - i for i, j in enumerate(lst))
    t = 0
    for i, els in groupby(pos):
        l = len(list(els))
        el = lst[t]
        t += l
        yield range(el, el+l)


def group_1(data):
    '''
    # Merge consecutive identical values
    :param data: pos index
    :return: start and end of consecutive same values
    '''
    last = data[0]
    start = end = 0
    for n in data[1:]:
        if n - last <= 5: # Part of the group, bump the end
            last = n
            end += 1
        else: # Not part of the group, yield current group and start a new
            yield range(data[start], data[end]+1)
            last = n
            start = end = end + 1
    # yield start, end
    yield range(data[start], data[end]+1)


def group(chrlist):
    '''
    # Merge consecutive identical chromosome IDs
    :param chrlist:
    :return:
    '''
    last = chrlist[0]
    start = end = 0
    for n in chrlist[1:]:
        if n == last: # Part of the group, bump the end
            last = n
            end += 1
        else: # Not part of the group, yield current group and start a new
            yield start, end
            last = n
            start = end = end + 1
    yield start, end


def group_position(pos1list, pos2list, binsize):
    result = defaultdict(list)
    l1 = sorted(set(pos1list), key=pos1list.index)
    pos1group = list(group_1(l1))
    for i in range(len(pos1group)):
        start_pos = pos1list.index(pos1group[i][0])
        end_pos = len(pos1list) - 1 - pos1list[::-1].index(pos1group[i][-1])
        cur_pos2_list = pos2list[start_pos:end_pos + 1]
        cur_pos1_list = pos1list[start_pos:end_pos + 1]
        cur_pos2_list_sort = sorted(cur_pos2_list)
        l2 = sorted(set(cur_pos2_list_sort), key=cur_pos2_list_sort.index)
        pos2group = list(group_1(l2))
        if len(pos2group) > 1:
            for j in range(len(pos2group)):
                new_pos1 = []
                for k in range(len(pos2group[j])):
                    if pos2group[j][k] in cur_pos2_list:
                        pos = cur_pos2_list.index(pos2group[j][k])
                        new_pos1.append(cur_pos1_list[pos])
                '''
                for m in range(len(pos1list)):
                    if pos2list[m] in pos2group[j]:
                        new_pos1.append(pos1list[m])
                
                for m in range(len(cur_pos1_list)):
                    if pos2list[m] in cur_pos2_list:
                        new_pos1.append(cur_pos1_list[m])
                '''
                new_pos1 = sorted(new_pos1)
                result['pos1_start'].append(new_pos1[0] * binsize)
                result['pos1_end'].append(new_pos1[-1] * binsize + binsize)
                result['pos2_start'].append(pos2group[j][0] * binsize)
                result['pos2_end'].append(pos2group[j][-1] * binsize + binsize)

        else:
            result['pos1_start'].append(pos1group[i][0] * binsize)
            result['pos1_end'].append(pos1group[i][-1] * binsize + binsize)
            result['pos2_start'].append(pos2group[0][0] * binsize)
            result['pos2_end'].append(pos2group[0][-1] * binsize + binsize)
    return result




import os, sys
import numpy as np
import cooler
import multiprocessing


def getBins(coolfile):
    """
     Read the header file of the cool file
    :param coolfile: input cool file
    :return: binsInfor {'chrom'：[start_pos, end_pos]}
    """

    binsInfo = {}
    chroms = coolfile.chroms()["name"][:]
    for chrom in chroms:
        idxarray = np.where(coolfile.bins()["chrom"][:] == chrom)
        chromstart = idxarray[0][0]
        chromend = idxarray[0][-1]
        binsInfo[chrom] = [chromstart, chromend]
    return binsInfo


def get_chromInfo(chromlf):

    """
    read reference genome length file
    :param chromlf: reference genome length file
    :return: chromID and chromosome length
    """
    chroms = []
    inf = open(chromlf)
    for line in inf:
        line = line.strip().split('\t')
        chroms.append(line[0])
    return chroms

def dumpMatrix(binSize, coolfile, binsInfo, chrom1, chrom2, outdir, name):
    """
    Convert cool file to contact matrix for each chromosome
    :param binSize: the resolution of Hi-C contact matrix
    :param coolfile: input cool file
    :param binsInfo: chromosome info in cool file
    :param chrom1: chrom1 id
    :param chrom2: chrom2 id
    :param outdir: path to the output file
    :param name: sample name
    :return: current Hi-C matrix
    """

    chrom1start, chrom1end = binsInfo[chrom1]
    chrom2start, chrom2end = binsInfo[chrom2]
    if 'chr' not in chrom1:
        outputname = os.path.join(outdir, name + '_%skb_%s_%s_matrix.txt' % (str(binSize // 1000),
                                                                             'chr'+chrom1, 'chr'+chrom2))
    else:
        outputname = os.path.join(outdir, name + '_%skb_%s_%s_matrix.txt' % (str(binSize // 1000), chrom1, chrom2))

    # matrixName = '_'.join([name, str(resolution)+'kb',chrom1,chrom2,"InterMap_matrix.txt"])
    # matrixfile = os.path.join(outdir,matrixName)
    # print(chrom1start, chrom1end, chrom2start, chrom2end)
    matrix = coolfile.matrix(balance=False)[chrom1start:(chrom1end + 1), chrom2start:(chrom2end + 1)]
    np.savetxt(outputname, matrix, fmt='%.2f', delimiter='\t')


def coolToMatrix(matrixFile, ref_file, binSize, outdir, name, n_cores):

    """
    :param matrixFile: input cool file
    :param binSize: the resolution of Hi-C contact matrix
    :param outdir: path to the output file
    :param name: sample name
    """

    coolfile = cooler.Cooler(matrixFile)
    binsInfo = getBins(coolfile)
    chr_id = get_chromInfo(ref_file)

    chr_count = len(chr_id)
    outdir_inter = os.path.join(outdir, "Inter_matrix")
    outdir_intra = os.path.join(outdir, "Intra_matrix")
    if not os.path.isdir(outdir_inter):
        os.mkdir(outdir_inter)
    if not os.path.isdir(outdir_intra):
        os.mkdir(outdir_intra)
    p = multiprocessing.Pool(processes=n_cores)

    for i in range(chr_count):
        for j in range(i, chr_count):
            if i == j:
                outdir = outdir_intra
            else:
                outdir = outdir_inter

            if 'chr' not in next(iter(binsInfo)):
                chrom1 = chr_id[i].lstrip('chr')
                chrom2 = chr_id[j].lstrip('chr')
            else:
                chrom1 = chr_id[i]
                chrom2 = chr_id[j]
            print('chroms: ', chrom1, chrom2)
            # dumpMatrix(binSize, coolfile, binsInfo, chrom1, chrom2, outdir, name)
            p.apply_async(dumpMatrix, args=(binSize, coolfile, binsInfo, chrom1, chrom2, outdir, name))
    p.close()
    p.join()




import os, pyBigWig
import numpy as np
import pandas as pd
from collections import Counter
from pygam import PoissonGAM, s


def get_contacts(matrixfile, chrID, binSize):
    '''
    Extract the coverage of each bin from the Hi-C matrix
    :param matrixfile: Hi-C matrix
    :param chrID: chrom ID
    :param binSize: resolution of the hic matrix
    :return: {'chrom', 'start', 'end', 'Coverage'}
    '''
    matrix = np.loadtxt(matrixfile)
    matrix = matrix[:-1, :-1]
    row, col = np.diag_indices_from(matrix)
    matrix[row, col] = 0
    data = pd.DataFrame(columns=['chrom', 'start', 'end', 'Coverage'])
    for k in range(len(matrix[0])):
        coverage = np.sum(matrix[k])
        start = k * binSize
        end = (k + 1) * binSize
        data = data.append({'chrom': chrID, 'start': start, 'end': end, 'Coverage': coverage}, ignore_index=True)
    return data


def signal_from_bigwig(table, bw_fil, name='GC'):
    '''
    Extract the features (GC and mappability) of each bin from the bigwig file
    :param table: chrom info {'chrom', 'start', 'end', 'Coverage'}
    :param bw_fil: bigwig file
    :param name: GC or mappability
    :return: {'chrom', 'start', 'end', 'Coverage', 'GC', 'mappability'}
    '''
    bw = pyBigWig.open(bw_fil)
    arr = []
    for i in table.index:
        row = table.loc[i]
        # (row['chrom'], row['start'], row['end'])
        v = bw.stats(row['chrom'], row['start'], row['end'], type='mean')[0]
        if v is None:
            v = 0
        arr.append(v)

    table[name] = np.r_[arr]

    return table


def count_REsites(table, npz_fil, res):
    '''
    Extract the REsites (GC and mappability) of each bin from the bigwig file
    :param table: chrom info {'chrom', 'start', 'end', 'Coverage'}
    :param npz_fil: RE file
    :param res: resolution of the hic matrix
    :return: {'chrom', 'start', 'end', 'Coverage', 'GC', 'mappability', 'RE'}
    '''
    RE = np.load(npz_fil)

    RE_by_bin = {}
    for chrom in RE:
        tmp = RE[chrom] // res
        RE_by_bin[chrom] = Counter(tmp)

    arr = []
    for i in table.index:
        row = table.loc[i]
        b_i = row['start'] // res
        chrom = row['chrom']
        arr.append(RE_by_bin[chrom][b_i])

    table['RE'] = np.r_[arr]

    return table


def filterZeros(table, cols=['GC', 'Mappability', 'RE']):
    '''
    Filter out bins with feature 0
    :param table: {'chrom', 'start', 'end', 'Coverage', 'GC', 'mappability', 'RE'}
    :param cols:
    :return: feature == 0 index and filtered tabel
    '''
    mask = (table['GC'] != 0) & (table['Mappability'] != 0) & (table['RE'] != 0) & (table['Coverage'] != 0)
    filtered = table[mask]

    return mask, filtered


def filterZeros_withoutRE(table, cols=['GC', 'Mappability']):
    mask = (table['GC'] != 0) & (table['Mappability'] != 0) & (table['Coverage'] != 0)
    filtered = table[mask]

    return mask, filtered


def normalized_coverage(matrixfile, gc_file, map_file, cutsize_file, binSize, chrID):
    # Normalized the coverage by possion GAM model
    table = get_contacts(matrixfile, chrID, binSize)
    table = signal_from_bigwig(table, gc_file, name='GC')
    # Load mappability scores
    table = signal_from_bigwig(table, map_file, name='Mappability')
    # Count the number of cut sizes for each bin
    table = count_REsites(table, cutsize_file, binSize)
    # Filter out invalid bins
    mask, filtered = filterZeros(table)
    # normalized coverage by generated addition model
    X = filtered[['GC', 'Mappability', 'RE']].values
    y = filtered['Coverage'].values
    gam = PoissonGAM(s(0) + s(1) + s(2), fit_intercept=True).fit(X, y)
    # gam.gridsearch(X, y)
    # logger.info('Output residuals ...')
    residuals = gam.deviance_residuals(X, y)
    residuals = np.array(residuals - residuals.min())
    # residuals = np.nan_to_num(residuals)
    idx = np.where(mask)[0]
    coverage = np.zeros(table.shape[0])
    coverage[idx] = residuals

    return coverage

def normalized_coverage_withoutRE(matrixfile, gc_file, map_file, binSize, chrID):
    # Normalized the coverage by possion GAM model
    table = get_contacts(matrixfile, chrID, binSize)
    table = signal_from_bigwig(table, gc_file, name='GC')
    # Load mappability scores
    table = signal_from_bigwig(table, map_file, name='Mappability')
    # Filter out invalid bins
    mask, filtered = filterZeros_withoutRE(table)
    # normalized coverage by generated addition model
    X = filtered[['GC', 'Mappability']].values
    y = filtered['Coverage'].values
    gam = PoissonGAM(s(0) + s(1), fit_intercept=True).fit(X, y)
    # logger.info('Output residuals ...')
    residuals = gam.deviance_residuals(X, y)
    residuals = np.array(residuals - residuals.min())
    # residuals = np.nan_to_num(residuals)
    idx = np.where(mask)[0]
    coverage = np.zeros(table.shape[0])
    coverage[idx] = residuals
    return coverage
'''
gc_file = "/mnt/d/HiSV_test_data/normData/hg19_1kb_GC.bw"
mapscore_file = "/mnt/d/HiSV_test_data/normData/hg19_mappability_100mer.1kb.bw"
cutsites_file = "/mnt/d/HiSV_test_data/normData/hg19.HindIII.npz"
# matrixfile = "/mnt/d/HiSV_test_data/test_data/hic_data/chr9.txt"
matrixfile = "/mnt/d/HiSV_test_data/result/Intra_matrix/K562_50kb_chr9_chr9_matrix.txt"
binSize = 50000
chrID = "chr9"
# table = get_contacts(matrixfile, chrID, binSize)
normed_coverage = normalized_coverage(matrixfile, gc_file, mapscore_file, cutsites_file, binSize, chrID)
'''




"""
Setup script for HiSV.

"""
import os, sys, glob
import setuptools

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

if (sys.version_info.major!=3) or (sys.version_info.minor!=7):
    print('PYTHON 3.7 IS REQUIRED. YOU ARE CURRENTLY USING PYTHON {}'.format(sys.version.split()[0]))
    sys.exit(2)

# Guarantee Unix Format
for src in glob.glob('scripts/*'):
    text = open(src, 'r').read().replace('\r\n', '\n')
    open(src, 'w').write(text)

setuptools.setup(
    name = 'hisv',
    version = "1.0.9",
    author = "Li Junping",
    author_email = 'lijunping02@qq.com',
    url = 'https://github.com/GaoLabXDU/HiSV',
    description = 'A computational pipeline for structural variation detection from Hi-C data',
    keywords = ("Hi-C", "structural variation", "control-free"),
    scripts = glob.glob('scripts/*'),
    packages = setuptools.find_packages(),
    include_package_data = True,
    platforms = "any",
    license="MIT Licence",
    install_requires = [
        "numpy==1.21.6",
        "pandas==1.3.5",
        "pyBigWig==0.3.22",
        "prox-tv==3.3.0",
        "pysam==0.21.0",
        "pytest==7.4.0",
        "pygam==0.8.0",
        "scikit-learn==1.0.2",
        "cooler==0.9.2",
        ]
    )



import os
import numpy as np
import pysam
import multiprocessing


def read_reflen(chr_len_file, binSize):
    """
    read chrom length file
    :param chr_len_file: reference genome length file
    :param binSize: the resolution of Hi-C contact matrix
    :return: chromID and the number of bins for each chromosome
    """
    with open(chr_len_file) as f:
        chr_id = []
        chr_bin_num = []
        for line in f:
            line_data = line.strip().split('\t')
            chr_id.append(line_data[0])
            chr_bin_num.append(int(int(line_data[1]) / binSize))

    chr_bin_num = np.array(chr_bin_num)
    return chr_id, chr_bin_num


def read_bam_file(filename, chrlist, matrix, binSize, outputname):
    """
    covert bam file to contact matrix for each chromosome
    :param filename: the bam file
    :param chrlist: chromID
    :param matrix: the initialized matrix
    :param binSize: the resolution of Hi-C contact matrix
    :return: the contact matrix for current chromosome
    """
    # covert bam file to contact matrix
    # return each chromosome matrix
    # pysam.index(filename)
    bam = pysam.AlignmentFile(filename, "rb")
    for read in bam:
        if read.reference_name == chrlist[0] and read.next_reference_name == chrlist[1]:
            row = int(read.reference_start / binSize)
            col = int(read.next_reference_start / binSize)
            matrix[row][col] += 1
        elif read.reference_name == chrlist[1] and read.next_reference_name == chrlist[0]:
            row = int(read.next_reference_start / binSize)
            col = int(read.reference_start / binSize)
            matrix[row][col] += 1
    np.savetxt(outputname, matrix, fmt='%.2f', delimiter='\t')


def bamToMatrix(bamfile, ref_file, binSize, outdir, name, n_cores):
    """

    :param bamfile: the bam file
    :param ref_file: reference genome length file
    :param binSize: the resolution of Hi-C contact matrix
    :param outdir: path to the output file
    :param name: sample name
    :param n_cores: cores number
    """
    chr_id, chr_bin_num = read_reflen(ref_file, binSize)
    chr_count = len(chr_id)
    outdir_inter = os.path.join(outdir, "Inter_matrix")
    outdir_intra = os.path.join(outdir, "Intra_matrix")
    if not os.path.isdir(outdir_inter):
        os.mkdir(outdir_inter)
    if not os.path.isdir(outdir_intra):
        os.mkdir(outdir_intra)
    # i, j : chri and chrj
    pysam.index(bamfile)
    p = multiprocessing.Pool(processes=n_cores)
    # i, chr_id, intra_hicfile, inter_hicfile, win, cutoff, name, binSize

    for i in range(chr_count):
        for j in range(i, chr_count):
            if i == j:
                outputname = os.path.join(outdir_intra, name + '_%skb_%s_%s_matrix.txt' % (str(binSize//1000),
                                                                                           chr_id[i], chr_id[j]))
            else:
                outputname = os.path.join(outdir_inter, name + '_%skb_%s_%s_matrix.txt' % (str(binSize//1000),
                                                                                           chr_id[i], chr_id[j]))
            print('chromosome:', chr_id[i], chr_id[j])
            raw_num = chr_bin_num[i] + 1
            col_num = chr_bin_num[j] + 1
            chrlist = [chr_id[i], chr_id[j]]
            # initialized matrix
            init_matrix = np.full((raw_num, col_num), 0)
            p.apply_async(read_bam_file, args=(bamfile, chrlist, init_matrix, binSize, outputname))
    p.close()
    p.join()




import os,sys
import numpy as np
import hicstraw
import multiprocessing


def get_chromInfo(chromlf):

    """
    read reference genome length file
    :param chromlf: reference genome length file
    :return: chromID and chromosome length
    """
    chroms = []
    infos = {}
    inf = open(chromlf)
    for line in inf:
        line = line.strip().split('\t')
        infos[line[0]] = int(line[1])
        chroms.append(line[0])
    return chroms, infos


def dumpMatrix(chrom1, chrom2, binSize, hicfile, chromInfo, outputMatrixFile):
    """
    convert hic file to matrix file for each chromosome
    :param chrom1: chrom1 ID
    :param chrom2: chrom2 ID
    :param binSize: the resolution of Hi-C contact matrix
    :param hicfile:  input hic file
    :param chromInfo: chromID and length
    :param outputMatrixFile: path to the output file
    """
    chrom1length = chromInfo[chrom1]
    chrom2length = chromInfo[chrom2]
    binnumber1 = int(np.divide(chrom1length, binSize)) + 1
    binnumber2 = int(np.divide(chrom2length, binSize)) + 1
    # Make chromID consistent with the definition of chromID in the hi-c file
    chr1 = chrom1.lstrip('chr')
    chr2 = chrom2.lstrip('chr')
    print("chromosome: ", chr1, chr2)
    # initialization contact matrix according the length of chromosomes
    init_matrix = np.full((binnumber1, binnumber2), 0)
    # result = hicstraw.straw('NONE', hicfile, str(chrom1), str(chrom2), 'BP', binSize)
    result = hicstraw.straw('observed', 'NONE', hicfile, str(chr1), str(chr2), 'BP', binSize)

    for i in range(len(result)):
        # print(result[i].binX, result[i].binY, result[i].counts)
        row = int(result[i].binX / binSize)
        col = int(result[i].binY / binSize)
        init_matrix[row][col] = result[i].counts

    if chrom1 == chrom2:
        init_matrix += init_matrix.T - np.diag(init_matrix.diagonal())
    np.savetxt(outputMatrixFile, init_matrix, fmt='%i', delimiter='\t')


def hicToMatrix(hicfile, binSize, ref_file, outdir, name, n_cores):
    """
    :param hicfile: input hic file
    :param binSize: the resolution of Hi-C contact matrix
    :param ref_file: reference genome length file
    :param outdir: path to the output file
    :param name: sample name
    :param n_cores: cores number
    """
    MatrixInfo = {}
    chroms, chromInfo = get_chromInfo(ref_file)
    chr_count = len(chroms)
    # Intra-chromosomal and inter-chromosomal contact matrices are stored separately
    outdir_inter = os.path.join(outdir, "Inter_matrix")
    outdir_intra = os.path.join(outdir, "Intra_matrix")
    if not os.path.isdir(outdir_inter):
        os.mkdir(outdir_inter)
    if not os.path.isdir(outdir_intra):
        os.mkdir(outdir_intra)
    p = multiprocessing.Pool(processes=n_cores)

    for i in range(chr_count):
        for j in range(i, chr_count):
            if i == j:
                outdir = outdir_intra
            else:
                outdir = outdir_inter

            chrom1 = chroms[i]
            chrom2 = chroms[j]
            outputname = os.path.join(outdir, name + '_%skb_%s_%s_matrix.txt' % (str(binSize//1000), chrom1, chrom2))
            # dumpMatrix(chrom1, chrom2, binSize, hicfile, chromInfo, outputname)
            p.apply_async(dumpMatrix, args=(chrom1, chrom2, binSize, hicfile, chromInfo, outputname))
    p.close()
    p.join()



